////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                            //
//                                       WATTO STUDIOS                                        //
//                             Java Code, Programs, and Software                              //
//                                    http://www.watto.org                                    //
//                                                                                            //
//                           Copyright (C) 2004-2010  WATTO Studios                           //
//                                                                                            //
// This program is free software; you can redistribute it and/or modify it under the terms of //
// the GNU General Public License published by the Free Software Foundation; either version 2 //
// of the License, or (at your option) any later versions. This program is distributed in the //
// hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranties //
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License //
// at http://www.gnu.org for more details. For updates and information about this program, go //
// to the WATTO Studios website at http://www.watto.org or email watto@watto.org . Thanks! :) //
//                                                                                            //
////////////////////////////////////////////////////////////////////////////////////////////////

package org.watto.android.component;

import android.view.ContextMenu.ContextMenuInfo;
import android.view.View;


/***********************************************************************************************
Provides additional information to a <code>View</code>, such as for use when displaying a
<code>ContextMenu</code> and you want to know which <code>View</code> is triggering it
***********************************************************************************************/
public class WSExtraInfo implements ContextMenuInfo {

  String[] keys;
  Object[] values;

  View parent;


  /***********************************************************************************************
  Creates a <code>WSExtraInfo</code> with no initial information
  ***********************************************************************************************/
  public WSExtraInfo(){
    keys = new String[0];
    values = new Object[0];
    parent = null;
  }


  /***********************************************************************************************
  Creates a <code>WSExtraInfo</code> populated with the input parameters
  @param keys the keys to identify each <i>value</i>
  @param values the <code>Object</code>s
  ***********************************************************************************************/
  public WSExtraInfo(String[] keys, Object[] values){
    this.keys = keys;
    this.values = values;
    parent = null;
  }


  /***********************************************************************************************
  Creates a <code>WSExtraInfo</code> with a single key <b>parent</b> set to the <i>view</i> that
  this <code>WSExtraInfo</code> represents
  @param view the <code>View</code> that this <code>WSExtraInfo</code> represents
  ***********************************************************************************************/
  public WSExtraInfo(View view){
    keys = new String[0];
    values = new Object[0];
    parent = view;
  }


  /***********************************************************************************************
  Creates a <code>WSExtraInfo</code> populated with the input parameters and a key <b>parent</b>
  set to the <i>view</i> that this <code>WSExtraInfo</code> represents
  @param view the <code>View</code> that this <code>WSExtraInfo</code> represents
  @param keys the keys to identify each <i>value</i>
  @param values the <code>Object</code>s
  ***********************************************************************************************/
  public WSExtraInfo(View view, String[] keys, Object[] values){
    this.keys = keys;
    this.values = values;
    this.parent = view;
  }


  /***********************************************************************************************
  Adds a single item of extra information
  @param key the key to identify the <i>value</i>
  @param value the <code>Object</code>
  ***********************************************************************************************/
  public void add(String key,Object value){
    int length = keys.length;
    resizeArrays(length + 1);

    keys[length] = key;
    values[length] = value;
  }


  /***********************************************************************************************
  Finds the position of the <i>key</i> in the <i>keys</i> array
  @param key the key to find
  @return the position in the <i>keys</i> array, or <b>-1</b> if the <i>key</i> wasn't found
  ***********************************************************************************************/
  public int findKeyPosition(String key){
    int length = keys.length;
    for (int i = 0;i < length;i++) {
      if (keys[i].equals(key)) {
        return i;
      }
    }
    return -1;
  }


  /***********************************************************************************************
  Gets the <code>Object</code> with the given <i>key</i>
  @param key the key of the <code>Object</code> to return
  @return the <code>Object</code> for the <i>key</i>, or <b>null</b> if the key doesn't exist
  ***********************************************************************************************/
  public Object get(String key){
    int arrayPosition = findKeyPosition(key);
    if (arrayPosition == -1) {
      return null;
    }
    else {
      return values[arrayPosition];
    }
  }


  /***********************************************************************************************
  Gets the <code>View</code> that this <code>WSExtraInfo</code> represents
  @return the <code>View</code> that this <code>WSExtraInfo</code> represents, or <b>null</b> if
  the parent doesn't exist or is not a <code>View</code>
  ***********************************************************************************************/
  public View getParent(){
    if (parent == null || !(parent instanceof View)) {
      return null;
    }
    return (View)parent;
  }


  /***********************************************************************************************
  Determines whether a value exists with the given <i>key</i> or not 
  @param key the key of the <code>Object</code> to find
  @return <b>true</b> if the <i>key</i> exists <br />
          <b>false</b> if the key is not defined
  ***********************************************************************************************/
  public boolean has(String key){
    int arrayPosition = findKeyPosition(key);
    if (arrayPosition == -1) {
      return false;
    }
    else {
      return true;
    }
  }


  /***********************************************************************************************
  Removes a single item of extra information
  @param key the key to identify the <i>value</i> to remove
  ***********************************************************************************************/
  public void remove(String key){
    int arrayPosition = findKeyPosition(key);
    if (arrayPosition == -1) {
      return;
    }
    else {
      // move the last item in the arrays over the top of the item to delete
      int length = keys.length - 1;
      keys[arrayPosition] = keys[length];
      values[arrayPosition] = values[length];

      // now shrink the arrays
      resizeArrays(length);
    }
  }


  /***********************************************************************************************
  Resizes the <i>keys</i> and <i>values</i> arrays
  @param length the new length of the arrays
  ***********************************************************************************************/
  public void resizeArrays(int length){
    // keys array
    int oldLength = keys.length;
    int copyLength = oldLength;
    if (length < copyLength) {
      copyLength = length;
    }

    String[] oldKeys = keys;
    keys = new String[length];
    System.arraycopy(oldKeys,0,keys,0,copyLength);

    // values array
    oldLength = values.length;
    copyLength = oldLength;
    if (length < copyLength) {
      copyLength = length;
    }

    Object[] oldValues = values;
    values = new Object[length];
    System.arraycopy(oldValues,0,values,0,copyLength);
  }


  /***********************************************************************************************
  Adds a single item of extra information, or replaces the value if the <i>key</i> already exists
  @param key the key to identify the <i>value</i>
  @param value the <code>Object</code>
  ***********************************************************************************************/
  public void set(String key,Object value){
    int arrayPosition = findKeyPosition(key);
    if (arrayPosition == -1) {
      add(key,value);
    }
    else {
      values[arrayPosition] = value;
    }
  }


  /***********************************************************************************************
  Sets the <code>View</code> that this <code>WSExtraInfo</code> represents
  @param view the <code>View</code> that this <code>WSExtraInfo</code> represents
  ***********************************************************************************************/
  public void setParent(View view){
    this.parent = view;
  }

}
