////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                            //
//                                       WATTO STUDIOS                                        //
//                             Java Code, Programs, and Software                              //
//                                    http://www.watto.org                                    //
//                                                                                            //
//                           Copyright (C) 2004-2010  WATTO Studios                           //
//                                                                                            //
// This program is free software; you can redistribute it and/or modify it under the terms of //
// the GNU General Public License published by the Free Software Foundation; either version 2 //
// of the License, or (at your option) any later versions. This program is distributed in the //
// hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranties //
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License //
// at http://www.gnu.org for more details. For updates and information about this program, go //
// to the WATTO Studios website at http://www.watto.org or email watto@watto.org . Thanks! :) //
//                                                                                            //
////////////////////////////////////////////////////////////////////////////////////////////////

package org.watto.android.component;

import android.content.Context;
import android.content.res.XmlResourceParser;
import android.util.AttributeSet;
import android.view.ContextMenu.ContextMenuInfo;
import android.widget.TextView;


/***********************************************************************************************
An extension of <code>TextView</code> that provides additional helper functions
***********************************************************************************************/
public class WSTextView extends TextView implements WSAndroidComponent {

  WSExtraInfo extraInfo;


  /***********************************************************************************************
  Constructor
  @param context the <code>Context</code> holding the <code>WSTextView</code>
  @param attributes the properties of this <code>WSTextView</code>, from the XML file
  ***********************************************************************************************/
  public WSTextView(Context context, AttributeSet attributes){
    super(context,attributes);
    extraInfo = new WSExtraInfo(this);

    scanAttributes(attributes);
  }


  /***********************************************************************************************
  Adds a single item of extra information
  @param key the key to identify the <i>value</i>
  @param value the <code>Object</code>
  ***********************************************************************************************/
  public void addExtraInfo(String key,Object value){
    extraInfo.add(key,value);
  }


  /***********************************************************************************************
  Gets the additional information associated with this <code>View</code>. Used when triggering a
  <code>ContextMenu</code> from this component.
  @return any additional information attached to this <code>View</code>
  ***********************************************************************************************/
  public ContextMenuInfo getContextMenuInfo(){
    return extraInfo;
  }


  /***********************************************************************************************
  Gets the additional information associated with this <code>View</code>.
  @return any additional information attached to this <code>View</code>
  ***********************************************************************************************/
  public WSExtraInfo getExtraInfo(){
    return extraInfo;
  }


  /***********************************************************************************************
  Gets the <code>Object</code> with the given <i>key</i>
  @param key the key of the <code>Object</code> to return
  @return the <code>Object</code> for the <i>key</i>, or <b>null</b> if the key doesn't exist
  ***********************************************************************************************/
  public Object getExtraInfo(String key){
    return extraInfo.get(key);
  }


  /***********************************************************************************************
  Determines whether there is extra information stored for the given <i>key</i>
  @param key the key of the <code>Object</code> to find
  @return <b>true</b> if extra information for the <i>key</i> exists <br />
          <b>false</b> if extra information for the key is not defined
  ***********************************************************************************************/
  public boolean hasExtraInfo(String key){
    return extraInfo.has(key);
  }


  /***********************************************************************************************
  Removes a single item of extra information
  @param key the key to identify the <i>value</i> to remove
  ***********************************************************************************************/
  public void removeExtraInfo(String key){
    extraInfo.remove(key);
  }


  /***********************************************************************************************
  Scans the <code>AttributeSet</code> for attributes in the <i>watto</i> namespace, then sets them
  in the <i>extraInfo</i> 
  @param attributes the properties of this <code>WSTextView</code>, from the XML file
  ***********************************************************************************************/
  public void scanAttributes(AttributeSet attributes){
    if (attributes instanceof XmlResourceParser) {
      XmlResourceParser xmlAttributes = (XmlResourceParser)attributes;
       
      String namespace = "http://www.watto.org/android/schema";

      int attributeCount = xmlAttributes.getAttributeCount();
      for (int i = 0;i < attributeCount;i++) {
        String prefix = xmlAttributes.getAttributeNamespace(i);
        if (prefix.equals(namespace)) {
          String name = xmlAttributes.getAttributeName(i);
          String value = xmlAttributes.getAttributeValue(i);
          extraInfo.add(name,value);
        }
      }
    }
  }


  /***********************************************************************************************
  Sets the additional information associated with this <code>View</code>.
  @param extraInfo any additional information for this <code>View</code>
  ***********************************************************************************************/
  public void setExtraInfo(WSExtraInfo extraInfo){
    this.extraInfo = extraInfo;
  }

}
